'''Curses window objects.'''

import types

from concepts.geometry import Position

from ui.properties import *



class Window:
    def __init__(self, ui, stdscr, model, rect, border = False):
        self.ui = ui
        self.stdscr = stdscr
        self.model = model
        self.rect = rect
        self.border = border
        self.window = stdscr.subwin(rect.height, rect.width, rect.y, rect.x)
        
    def can_write(self, pos, string = ' '):
        if pos.y < 0 or pos.y >= self.rect.height - self.border * 2:
            return False
        
        if pos.x < 0 and abs(pos.x) >= len(string):
            return False

        if self.rect.width - pos.x - self.border * 2 <= 0:
            return False
        
        return True
        
    def write(self, pos, string, attributes = []): 
        if not self.can_write(pos) or string is None:
            return
        
        if pos.x < 0:
            self.write(Position(0, pos.y), string[abs(pos.x):])
            return
        
        max_len = self.rect.width - pos.x - self.border * 2

        if len(string) > max_len:
            string = string[:max_len]
        
        x = pos.x + self.border
        y = pos.y + self.border
        
        try:
            if attributes:
                self.window.addstr(y, x, string, reduce(lambda a, b: a | b,
                                                        attributes))
            else:
                self.window.addstr(y, x, string)
        except:
            # Curses throws an exception after writing the bottom right
            # character of a window... ignoring it seems to work
            pass
        
    def clear(self):
        self.window.clear()
        
        if self.border:
            self.window.box()
    
    def erase(self):
        self.window.erase()
        
        if self.border:
            self.window.box()
        
    def refresh(self):
        self.window.refresh()
    
    def sync(self):
        self.window.syncup()
        
    def display(self):
        raise 'Not implemented'



class ActionWindow(Window):
    def display(self):
        self.clear()
        self.display_actions(0)
        
    def display_actions(self, line):
        displayed = []
        
        for action in self.model.available_actions():
            if not action.can_do() and not action.always_known():
                continue
            
            group = self.ui.group_for(action)
            
            if group is None:
                key  = self.ui.char_for(action)
                name = str(action)
                info = action.info()
            else:
                (key, name) = group
                info = ''
                
            if key in displayed:
                continue
            
            displayed += [key]
            
            if key is not None and action.can_do():
                start = 3 - len(key)
                end   = start + len(key) + 1
                
                self.write(Position(start,     line), '[', [BLACK])
                self.write(Position(start + 1, line), key, [WHITE])
                self.write(Position(end,       line), ']', [BLACK])
            
            
            self.write(Position(6, line), name, [GREY])
            self.write(Position(self.rect.width - self.border * 2 - len(info),
                                line), info, [GREY])
            
            line += 1



class MapWindow(Window):
    def __init__(self, ui, stdscr, model, rect, border = False):
        Window.__init__(self, ui, stdscr, model, rect, border)
        
        self._vector = None
        
    def display(self):
        self.erase()
        self.draw(self.map(), self.units())
        
    def draw(self, map, units):
        self._vector = None
        vector = self.vector()
        start  = -vector.dy - 1
        stop   = start + self.rect.height
        
        if start < 0:
            self._vector.dy += start + 1
            stop -= start
            start = 0
        
        if stop > map.height:
            self._vector.dy += stop - map.height - 1
            start -= stop - map.height
            start  = 0
            stop   = map.height
        
        for row in range(start, stop):
            for tile in map.reverse[row]:
                self.draw_tile(tile)
        
        for unit in units:
            self.draw_unit(unit)
                        
    def top_left(self):
        diagonal = self.rect.diagonal()
        diagonal.invert()
        diagonal.descale(2)
        
        pos = self.center()
        pos.translate(diagonal)
        
        return pos
    
    def vector(self):
        if self._vector is None:
            self._vector = self.top_left().direction()
            self._vector.invert()
            self._vector.dx  = 0
        
        return self._vector
    
    def draw_unit(self, unit):
        pos = unit.position()
        pos.translate(self.vector())
        
        if self.can_write(pos):
            self.write(pos, self.char_for_unit(unit), 
                       [self.color_for_unit(unit)])
        
    def draw_tile(self, tile):
        pos = tile.position()
        pos.translate(self.vector())

        if self.can_write(pos):
            self.write(pos, self.char_for_tile(tile), 
                       [self.color_for_tile(tile)])
    